class TriflError(Exception):
    """Base class for exceptions in this project."""

class DtaFileNotFound(TriflError):
    """Raised when a DTASelect file is not found at the expected path."""

class CimageFlatFileNotFound(TriflError):
    """Raised when a cimage flatfile (output_to_excel.txt) is not found at the expected path."""

class FilterNotFoundException(TriflError):
    """Raised when filter is not found."""



from trifl.parse import (
    parse_census,
    parse_dtaselect,
    parse_cimage_flatfile
)
import trifl.utils as utils
import trifl.config as config
import trifl.exceptions as exceptions

from Bio import SwissProt
from playhouse.sqlite_ext import (
    SqliteExtDatabase,
    JSONField,
)
from peewee import *

import datetime
import hashlib
import pathlib
import json


database = SqliteExtDatabase(None)

def create_tables():
    with database:
        database.create_tables([
            AnalysisModel,
            Blacklist,
            Whitelist,
            FastaDatabase,
            FastaDatabaseIndex,
            SilacExperiment,
            IsotopExperiment,
            TmtExperiment,
            TmtIsotopExperiment,
            SilacData,
            IsotopData,
            TmtData,
            TmtIsotopData,
        ])


class BaseModel(Model):
    class Meta:
        database = database
        legacy_table_names = False

class AnalysisModel(BaseModel):
    """Data on analyses of multiple datasets."""
    name = TextField()
    # analysis_hash = TextField(default=utils.get_git_revision_hash)
    analysis_hash = TextField(default='')
    date = DateField(default=datetime.datetime.now)
    datasets = JSONField()
    filters = JSONField(json_dumps=lambda x: json.dumps(x, cls=utils.JSONSetEncoder), null=True)

class BaseExperiment(BaseModel):
    """Holds experimental metadata."""
    name = TextField()
    date = DateField(default=datetime.datetime.now)
    meta = JSONField(null=True)

class SilacExperiment(BaseExperiment):
    """Holds experimental metadata."""
    file_hash = TextField(unique=True)
    dta_file_hash = TextField(unique=True)
    cimage_folder_name = TextField()
    database = DeferredForeignKey('FastaDatabaseIndex')

    @classmethod
    def get_dta_hash(cls, cimage_folder_path: pathlib.Path, dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME) -> str:
        dta_file_path = cls.get_dta_file_in_folder(cimage_folder_path, dta_folder_name)
        return utils.get_file_hash(dta_file_path)

    @classmethod
    def get_file_hash(cls, cimage_folder_path: pathlib.Path, parser: str, dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME) -> str:
        if parser == 'flatfile':
            file_path = cls.get_flatfile_in_folder(cimage_folder_path, dta_folder_name)
        elif parser in ('protein', 'peptide'):
            file_path = cimage_folder_path.joinpath(f'combined_{dta_folder_name}.txt')

        return utils.get_file_hash(file_path)

    @classmethod
    def get_by_path_hash(cls, cimage_folder_path: pathlib.Path, parser: str, dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME):
        dta_file_hash = cls.get_dta_hash(cimage_folder_path, dta_folder_name)
        file_hash = cls.get_file_hash(cimage_folder_path, parser, dta_folder_name)

        return cls.get_or_none(
            (cls.file_hash == file_hash) |
            (cls.dta_file_hash == dta_file_hash)
        )

    @classmethod
    def get_flatfile_in_folder(cls, cimage_folder_path: pathlib.Path, dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME):
        search_glob = 'output_*.to_excel.txt'
        flatfile_path = next(cimage_folder_path.glob(search_glob), None)
    
        if flatfile_path is None:
            raise exceptions.CimageFlatFileNotFound(
                f'Cimage flatfile not found in folder {cimage_folder_path}, using pattern {search_glob}'
            )

        return flatfile_path

    @classmethod
    def get_dta_file_in_folder(cls, cimage_folder_path: pathlib.Path, dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME):
        search_glob = 'DTASelect-filter*.txt'
        dta_file_path = next(cimage_folder_path.glob(search_glob), None)

        if dta_file_path is None:
            raise exceptions.DtaFileNotFound(
                f'DTASelect file not found in folder {cimage_folder_path}, using pattern {search_glob}'
            )

        return dta_file_path

    @classmethod
    def get_or_create(cls,
        name: str,
        cimage_folder_path: pathlib.Path,
        fasta_db_path: pathlib.Path,
        base_url: str,
        parser: str = config.DEFAULT_CIMAGE_PARSER,
        dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME,
    ):
        create_methods = {
            'flatfile': cls.create_from_flatfile
        }

        existing = None

        try:
            existing = cls.get_by_path_hash(cimage_folder_path, parser, dta_folder_name)
        except (exceptions.CimageFlatFileNotFound, exceptions.DtaFileNotFound):
            utils.download_cimage(
                base_url,
                cimage_folder_path,
                # cimage_folder_path.name,
                str(cimage_folder_path.relative_to(cimage_folder_path.parts[0])),
                dta_folder_name
            )

        if existing is not None:
            return existing

        return create_methods[parser](
            name=name,
            cimage_folder_path=cimage_folder_path,
            fasta_db_path=fasta_db_path,
            dta_folder_name=dta_folder_name
        )

    @classmethod
    def create_from_flatfile(cls,
        name: str,
        cimage_folder_path: pathlib.Path,
        fasta_db_path: pathlib.Path,
        dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME
    ):
        cimage_folder_path = pathlib.Path(cimage_folder_path)

        flatfile_path = cls.get_flatfile_in_folder(cimage_folder_path, dta_folder_name)
        flatfile_hash = utils.get_file_hash(flatfile_path)

        dta_file_path = cls.get_dta_file_in_folder(cimage_folder_path, dta_folder_name)
        dta_file_hash = utils.get_file_hash(dta_file_path)

        fasta_db = FastaDatabaseIndex.get_or_create(fasta_db_path)

        data = parse_cimage_flatfile(
            path=flatfile_path,
            dta_file_path=dta_file_path,
            fasta_db_path=fasta_db_path
        )

        with database.atomic():
            experiment = SilacExperiment.create(
                name=name,
                file_hash=flatfile_hash,
                dta_file_hash=dta_file_hash,
                cimage_folder_name=cimage_folder_path.name,
                database=fasta_db
            )
            SilacData.bulk_create([SilacData(**d, experiment=experiment) for d in data], batch_size=1000)
            return experiment

class IsotopExperiment(BaseExperiment):
    """Hold information about individual experiments."""
    file_hash = TextField()
    reactive_residue = FixedCharField(default=config.DEFAULT_REACTIVE_RESIDUE, max_length=1)

class TmtExperiment(BaseExperiment):
    """Hold information about individual experiments."""
    file_hash = TextField(unique=True)
    file_path = TextField()

    @classmethod
    def get_by_file_hash(cls, path):
        file_hash = utils.get_file_hash(path)
        return cls.get_or_none(cls.file_hash == file_hash)

    @classmethod
    def create_from_census(cls,
        name: str,
        census_file_path: pathlib.Path,
        fasta_db_path: pathlib.Path
    ):
        census_file_path = pathlib.Path(census_file_path)
        file_hash = utils.get_file_hash(census_file_path)

        data = parse_census(census_file_path, fasta_db_path)

        with database.atomic():
            experiment = TmtExperiment.create(
                name=name,
                file_hash=file_hash,
                file_path=census_file_path.name,
            )
            TmtData.bulk_create([TmtData(**d, experiment=experiment) for d in data], batch_size=1000)
            return experiment


class TmtIsotopExperiment(TmtExperiment):
    """Hold information about individual experiments."""
    reactive_residue = FixedCharField(default=config.DEFAULT_REACTIVE_RESIDUE, max_length=1)

class BaseData(BaseModel):
    """Holds actual experimental data."""
    experiment = ForeignKeyField(BaseExperiment) # this needs to be overriden if using as base class
    uniprot = TextField(index=True)
    description = TextField()
    symbol = TextField(index=True)
    sequence = TextField(index=True)
    clean_sequence = TextField(index=True)
    unique_sequence = BooleanField(index=True)
    meta = JSONField(null=True)

class IsotopData(BaseData):
    """Holds actual experimental data."""
    cimage_sequence = TextField(index=True)
    residue = IntegerField(index=True)
    possible_residues = JSONField(index=True)
    mass = FloatField()
    ratio = FloatField(index=True)
    stats = TextField()
    run = IntegerField()
    charge = IntegerField()
    segment = IntegerField()
    num_ms2 = IntegerField(null=True)
    link = TextField()

class SilacData(BaseData):
    """Holds actual experimental data."""
    mass = FloatField()
    ratio = FloatField(index=True)
    rsquared = FloatField(index=True)
    charge = IntegerField(index=True)
    segment = IntegerField(index=True)
    num_ms2 = IntegerField(index=True)

class TmtData(BaseData):
    experiment = ForeignKeyField(TmtExperiment)
    channel_intensities = JSONField()
    scan_num = IntegerField(index=True)
    tmt_purity = FloatField()
    filename = TextField(index=True)

    @property
    def control_intensities(self):
        '''Note this is actually set in during the analysis. This is a dummy property to document this behavior.'''

class TmtIsotopData(TmtData):
    """Holds actual experimental data."""
    experiment = ForeignKeyField(TmtIsotopExperiment)
    residue = IntegerField(index=True)
    possible_residues = JSONField(index=True)

class Blacklist(BaseModel):
    """List of entries to not consider in analyses."""
    data = ForeignKeyField(BaseData)
    description = TextField(null=True)

    class Meta:
        indexes = ((('data_id', 'description'), True),)

class Whitelist(BaseModel):
    """List of entries to consider in analyses irrespective of filters."""
    data = ForeignKeyField(BaseData)
    description = TextField(null=True)
    filter_subset = TextField(null=True)

    class Meta:
        indexes = ((('data_id', 'description'), True),)

class FastaDatabaseIndex(BaseModel):
    """FASTA database index."""
    file_name = TextField(unique=True, index=True)
    file_hash = TextField(unique=True)
    description = TextField(null=True)

    @classmethod
    def get_or_create(cls, path: pathlib.Path, description: str = ''):
        db_path = pathlib.Path(path)
        file_hash = utils.get_file_hash(db_path)

        try:
            with database.atomic():
                db = FastaDatabaseIndex.create(
                    file_name=db_path.name,
                    file_hash=utils.get_file_hash(path),
                    description=description
                )

                FastaDatabase.from_file(db_path, db)

                return db

        except IntegrityError:
            db = FastaDatabaseIndex.get(
                file_name=db_path.name,
                file_hash=file_hash
            )

            return db

    def is_sequence_unique(self, sequence: str) -> bool:
        count = (FastaDatabase
            .select(fn.Count(FastaDatabase.id))
            .where(FastaDatabase.sequence.contains(sequence))
        ).get()

        assert count != 0, 'Sequence not found in FASTA database!'
        return count == 1


class FastaDatabase(BaseModel):
    """Contains a uniprot database used for searching mass spec data."""
    database = ForeignKeyField(FastaDatabaseIndex)
    uniprot = TextField(unique=True, index=True)
    description = TextField(unique=True, index=True)
    sequence = TextField(index=True)

    @classmethod
    def from_file(cls, path: pathlib.Path, database: FastaDatabaseIndex):
        uniprot_db = utils.uniprot_db_from_fasta(path)

        db_models = [
            FastaDatabase(
                database=database,
                uniprot=uniprot,
                description=entry.description,
                sequence=str(entry.seq),
            )
            for uniprot, entry
            in uniprot_db.items()
        ]

        FastaDatabase.bulk_create(db_models, batch_size=250)

class UniprotFlatfile(Model):
    """Contains data from uniprot flatfile."""
    primary_accession = TextField(unique=True, index=True)
    primary_tax_id = IntegerField(index=True)

    # these are dumped directly from data
    gene_name = TextField(index=True)
    organism = TextField(index=True)
    description = TextField(index=True)
    sequence = TextField()
    comments = JSONField()
    cross_references = JSONField()

    # these too but they are more secondary right now
    accessions = JSONField()
    annotation_update = JSONField()
    created = JSONField()
    data_class = TextField()
    entry_name = TextField()
    features = JSONField()
    host_organism = JSONField()
    host_taxonomy_id = JSONField()
    keywords = JSONField()
    molecule_type = JSONField(null=True)
    organelle = JSONField()
    organism_classification = JSONField()
    protein_existence = IntegerField()
    references = JSONField()
    seqinfo = JSONField()
    sequence_length = IntegerField()
    sequence_update = JSONField()
    taxonomy_id = JSONField()

    @classmethod
    def from_file(cls, path):
        db_records = []

        with open(path) as f:
            db = SwissProt.parse(f)

            for record in db:
                db_records.append(cls(
                    primary_accession=record.accessions[0],
                    primary_tax_id=int(record.taxonomy_id[0]),
                    gene_name=record.gene_name,
                    organism=record.organism,
                    description=record.description,
                    sequence=record.sequence,
                    comments=record.comments,
                    cross_references=record.cross_references,
                    accessions=record.accessions,
                    annotation_update=record.annotation_update,
                    created=record.created,
                    data_class=record.data_class,
                    entry_name=record.entry_name,
                    features=record.features,
                    host_organism=record.host_organism,
                    host_taxonomy_id=record.host_taxonomy_id,
                    keywords=record.keywords,
                    molecule_type=record.molecule_type,
                    organelle=record.organelle,
                    organism_classification=record.organism_classification,
                    protein_existence=record.protein_existence,
                    references=[x.__dict__ for x in record.references],
                    seqinfo=record.seqinfo,
                    sequence_length=record.sequence_length,
                    sequence_update=record.sequence_update,
                    taxonomy_id=record.taxonomy_id,
                ))

        with database:
            database.create_tables([cls])

        with database.atomic():
            cls.bulk_create(db_records, batch_size=250)



from .shared import *
from .shared import filter_fun as f, Requirement as r

from typing import List
from copy import deepcopy
import statistics

RatioList = List[float]

@f(requires='ratio', level='individual')
def is_quantified(ratio: float) -> bool:
    return ratio > 0

@f(requires=r(['ratio', 'num_ms2'], ['ratio', 'num_ms2']), level='individual')
def is_not_singleton(ratio: float, num_ms2: int) -> bool:
    return ratio != 20 or num_ms2 > 1

@f(requires='rsquared', level='individual')
def passes_rsquared_cutoff(rsquared: float, min_rsquared: float = 0.5):
    return rsquared >= min_rsquared

@f(requires='ratio', level=['experiment', 'aggregate'])
def filter_20s(ratios: RatioList, ratio_cutoff: int = 2) -> List[bool]:
    """Filter out erroneous seeming 20s."""
    # 20s are stripped if the following conditions are met:
    # - the set of ratios is not just composed of 0s and 20s
    # - there is only one 20
    # - the lowest non-zero, non-20 value is below a cutoff
    no_filter = [True] * len(ratios)
    if 20 not in ratios or ratios.count(20) > 1:
        return no_filter

    only_zeros_and_20s = len([x for x in ratios if x > 0 and x != 20]) == 0
    min_below_cutoff = min(ratios) < ratio_cutoff

    if not only_zeros_and_20s and min_below_cutoff:
        return [r != 20 for r in ratios]
    else:
        return no_filter


@f(requires='ratio', level=['experiment', 'aggregate'])
def filter_two_or_more_20s(ratios: RatioList, ratio_cutoff=2) -> List[bool]:
    """Filter out erroneous seeming 20s when there is more than one 20 value."""
    # 20s are stripped if the following conditions are met:
    # - the set of ratios is not just composed of 0s and 20s
    # - there are at least 2 20 ratios
    # - the lowest non-zero, non-20 value is below a cutoff
    no_filter = [True] * len(ratios)
    if 20 not in ratios or ratios.count(20) < 2:
        return no_filter

    only_zeros_and_20s = len([x for x in ratios if x > 0 and x != 20]) == 0
    min_below_cutoff = min(ratios) < ratio_cutoff

    if not only_zeros_and_20s and min_below_cutoff:
        return [r != 20 for r in ratios]
    else:
        return no_filter

@f(requires='ratio', level=['experiment', 'aggregate'])
def filter_by_stdev(ratios: RatioList, stdev_cutoff: float = 0.6, ratio_cutoff: float = 4.0) -> List[bool]:
    """Filter ratios based on a standard deviation cutoff."""
    no_filter = [True] * len(ratios)
    quantified_ratios = [r for r in ratios if r > 0]

    # no-op if less than two ratios
    if len(quantified_ratios) < 2:
        return no_filter

    stdev = statistics.stdev(quantified_ratios)
    mean = statistics.mean(quantified_ratios)
    minimum = min(quantified_ratios)

    # if stdev is tight enough relative to the mean
    # or if all the ratios are above a certain threshold
    # return unchanged
    # else, set everything but the minimum to zero
    if (stdev != None and stdev / mean < stdev_cutoff) or minimum > ratio_cutoff:
        return no_filter
    else:
        return [r == minimum for r in ratios]

@f(requires='ratio', level=['experiment', 'aggregate'])
def filter_20s_by_stdev(ratios: RatioList, stdev_cutoff: float = 0.6, ratio_cutoff: float = 4) -> List[bool]:
    """Filter out 20s from sets of ratios with high standard deviation."""
    no_filter = [True] * len(ratios)
    quantified_ratios = [r for r in ratios if r > 0]

    # require that there are at least two ratios
    # that there is a 20 in the set of ratios
    # and that the minimum ratio is below the ratio_cutoff
    # otherwise do nothing
    if len(quantified_ratios) < 2 or 20 not in quantified_ratios or min(quantified_ratios) > ratio_cutoff:
        return no_filter

    above_threshold = statistics.stdev(quantified_ratios) / statistics.mean(quantified_ratios) > stdev_cutoff

    if above_threshold:
        return [r != 20 for r in ratios]
    else:
        return no_filter


@f(requires='ratio', level='aggregate')
def min_num_datasets(ratios: RatioList, min_datasets: int = 2) -> List[bool]:
    no_filter = [True] * len(ratios)
    filter_all = [False] * len(ratios)
    quantified_ratios = len([x for x in ratios if x > 0])
    return no_filter if quantified_ratios >= min_datasets else filter_all

@f(requires='clean_sequence', level=['experiment', 'aggregate'])
def min_num_peptides(clean_sequences: List[str], min_peptides: int=2) -> bool:
    no_filter = [True] * len(clean_sequences)
    filter_all = [False] * len(clean_sequences)
    return no_filter if len(list(utils.flatten(clean_sequences))) >= min_peptides else filter_all



from .shared import *
from .shared import filter_fun as f, Requirement as r

import numpy as np
from scipy import stats as stats
from typing import List

ChannelIntensities = List[int]


@f(requires=r('channel_intensities', 'control_intensities'), level='individual')
def control_sum(control_intensities: ChannelIntensities, min_intensity_per_control_channel: int = 5000) -> bool:
    min_control_sum = len(control_intensities) * min_intensity_per_control_channel
    control_sum = sum(control_intensities)
    return control_sum >= min_control_sum

@f(requires=r('channel_intensities', 'control_intensities'), level='individual')
def control_cv(control_intensities: ChannelIntensities, max_control_cv: float = 0.5) -> bool:
    control_cv = stats.variation(control_intensities)
    return control_cv <= max_control_cv

@f(requires=r('channel_intensities', 'channel_intensities'), level=['individual', 'experiment', 'aggregate'])
def max_cv(channel_intensities: ChannelIntensities, max_cv: float = 0.5) -> bool:
    cv = stats.variation(channel_intensities)
    return cv <= max_cv

@f(requires=r('channel_intensities', 'channel_intensities'), level=['individual'])
def max_condition_cv(channel_intensities: ChannelIntensities, channel_indices: List[int], max_cv: float = 0.5):
    cv = stats.variation(channel_intensities.T[:, channel_indices].T)
    return cv <= max_cv

@f(requires=r('channel_intensities', 'channel_intensities'), level=['individual'])
def channel_is_nonzero(channel_intensities: ChannelIntensities, channel_numbers: List[int]):
    return (channel_intensities.T[:, np.array(channel_numbers) - 1] != 0).all(axis=1)

@f(requires=r('channel_intensities', 'channel_intensities'), level=['individual'])
def total_intensity_sum(channel_intensities: ChannelIntensities, min_total_intensity_sum: int):
    intensity_sum = sum(channel_intensities)
    return intensity_sum >= min_total_intensity_sum

@f(requires=r('tmt_purity', 'tmt_purity'), level=['individual'])
def tmt_purity(tmt_purity: float, min_tmt_purity: float = 0.5):
    return tmt_purity > min_tmt_purity

# @f(requires=r('channel_intensities', 'channel_intensities'), level='aggregate')
# def min_number_of_datasets(channel_intensities: ChannelIntensities, max_cv: float = 0.5) -> bool:
#     cv = stats.variation(channel_intensities)
#     return cv <= max_cv




from trifl.models import (
    database,
    create_tables,
)
from trifl.dataset import (
    TmtDataset,
    SilacDataset
)
import trifl.filters.isobaric as isobaric_filters
import trifl.filters.isotopic as isotopic_filters
import trifl.config as config
import trifl.models as m
import trifl.utils as utils

# installed modules
import pandas as pd

# base modules
from dataclasses import dataclass
from typing import Optional
from functools import partial, reduce
from collections import defaultdict
from collections.abc import Iterable
from abc import ABC, abstractmethod
import itertools
import dataclasses
import pathlib
import operator



@dataclass
class AnalysisParams:
    name: Optional[str] = None
    datasets: Optional[dict] = None
    filters: Optional[dict] = None
    fasta_db_path: Optional[str] = None
    datasets_base_path: Optional[str] = None
    output_folder: Optional[str] = None

    @classmethod
    def from_dict(cls, dct):
        return cls(**{f.name: dct[f.name] for f in dataclasses.fields(cls)})


@dataclass
class SilacAnalysisParams(AnalysisParams):
    parser: Optional[str] = config.DEFAULT_CIMAGE_PARSER
    combine: Optional[str] = config.DEFAULT_SILAC_COMBINE_LEVEL
    datasets_base_url: Optional[str] = ''
    dta_folder_name: Optional[str] = config.DEFAULT_DTA_FOLDER_NAME


class _Analysis(ABC):
    def __init__(self, params: AnalysisParams) -> None:
        # create all tables if they do not already exist
        create_tables()

        self.params = params
        self.name = params.name
        self.filters = self.init_filters(params.filters)
        self.fasta_db_path = pathlib.Path(params.fasta_db_path)
        self.output_path = pathlib.Path(params.output_folder)

        self.datasets = self.init_datasets(
            params.datasets,
            base_path=pathlib.Path(params.datasets_base_path)
        )

        self.fasta_db = m.FastaDatabaseIndex.get_or_create(self.fasta_db_path)
        self.results = None

    @property
    def filtered_out(self) -> set:
        filtered = set()

        for filtered_from_dataset in self._analysis.filters.values():
            for filters in filtered_from_dataset.values():
                for filtered_ids in filters.values():
                    filtered.update(filtered_ids)

        return filtered

    def _new_analysis(self):
        return m.AnalysisModel(
            name=self.name,
            datasets=[d.as_dict() for d in self.datasets]
        )

    def run(self):
        analysis = self._new_analysis()
        results, filtered_out = self.apply_filters()
        analysis.filters = filtered_out
        analysis.save()
        self._analysis = analysis
        database.close()

    def init_filters(self, filters_dict: dict):
        filter_objects = defaultdict(list)

        for level, fs in filters_dict.items():
            if fs is None:
                continue

            for f in fs:
                filter_objects[level].append(self.filter_module.Filter.from_dict(f, self.filter_module))

        return filter_objects

    def apply_filters(self):
        result = dict()

        for dataset in self.datasets:
            _, filtered_out = dataset.apply_filters()
            result[dataset.name] = filtered_out

        return (None, result)

    @property
    def experiments_included(self):
        return list(itertools.chain.from_iterable(d.experiments for d in self.datasets))

    @property
    def experiment_ids_included(self):
        return [e.id for e in self.experiments_included]

class TmtAnalysis(_Analysis):
    data_model = m.TmtData
    filter_module = isobaric_filters
    dataset_class = TmtDataset

    def init_datasets(self, datasets: dict, base_path=''):
        return [TmtDataset(
            name=dataset['name'],
            filters=self.filters,
            replicate_paths=[pathlib.Path(base_path).joinpath(r) for r in dataset['replicates']],
            channel_layout=dataset['channel_layout'],
            control_channels=dataset['control_channels'],
            fasta_db_path=self.fasta_db_path
        ) for dataset in datasets]

    def report(self, report_prefix=config.DEFAULT_REPORT_PREFIX):
        m = self.data_model
        d = self.datasets[0]

        query = (m
            .select(
                m.id,
                m.experiment,
                m.filename,
                m.scan_num,
                m.uniprot,
                m.symbol,
                m.description,
                m.unique_sequence,
                m.clean_sequence,
                m.channel_intensities,
            )
            .where(
                (m.experiment_id.in_(self.experiment_ids_included))
            )
        )

        df = pd.DataFrame.from_records(list(query.dicts()))
        df = df.set_index('id')

        for channel in d.channel_layout:
            channel_num, channel_name = next(iter(channel.items()))
            new_col_name = f'{channel_name}.intensity'
            # append number to column to distinguish duplicates
            new_col_name = f'{new_col_name}_{len(df.filter(regex=new_col_name).columns)}'
            df[new_col_name] = df.channel_intensities.str.get(channel_num - 1)

        df = df.drop(columns=['channel_intensities'])

        control_cols = df.filter(regex=f'{d.control_channels}.intensity')
        data_cols = df.filter(regex=f'.+\.intensity')

        percentages = data_cols.div(control_cols.mean(axis=1), axis=0).mul(100, axis=0)
        percentages = percentages.rename(columns={
            c: c.replace('intensity', 'percent_of_control')
            for c in df.filter(regex='\.intensity').columns
        })

        df = df.join(percentages)

        report_output_name_template = '{}_{}_{}_{{}}.csv'.format(
            report_prefix,
            self.name,
            self._analysis.id
        )

        report_output_path = utils.get_timestamped_report_path(
            report_output_name_template,
            self.output_path
        )

        result_df.to_csv(report_output_path, index=False, encoding='utf-8-sig')
        return result_df

    def filter_report(self, filename: str = ''):
        pass


class SilacAnalysis(_Analysis):
    data_model = m.SilacData
    filter_module = isotopic_filters
    dataset_class = SilacDataset

    def init_datasets(self, datasets: dict, base_path=''):
        return [self.dataset_class(
            name=dataset['name'],
            filters=self.filters,
            replicate_paths=[pathlib.Path(base_path).joinpath(r) for r in dataset['replicates']],
            fasta_db_path=self.fasta_db_path,
            parser=self.params.parser,
            base_url=self.params.datasets_base_url,
            combine_level=self.params.combine,
            dta_folder_name=dataset.get('dta_folder_name', self.params.dta_folder_name),
        ) for dataset in datasets]

    def report(self, report_prefix=config.DEFAULT_REPORT_PREFIX, whitelist=None, blacklist=None):
        m = self.data_model

        expression_list = [
            m.experiment_id.in_(self.experiment_ids_included),
            m.id.not_in(self.filtered_out),
        ]

        if whitelist:
            expression_list.append(m.uniprot.in_(whitelist))

        if blacklist:
            expression_list.append(m.uniprot.not_in(blacklist))

        where_expression = reduce(operator.and_, expression_list)

        query = (m
            .select(
                m.id,
                m.experiment,
                m.uniprot,
                m.symbol,
                m.ratio,
            )
            .where(where_expression)
        )

        df = pd.DataFrame.from_records(list(query.dicts()))
        df = self.dataset_class.generate_id(df)

        for dataset in self.datasets:
            df.loc[df.experiment.isin(dataset.experiment_ids_included), 'condition'] = dataset.name

        df = df.set_index('id')

        result_df = (df
            .groupby(['uniprot', 'symbol', 'condition', 'experiment'])
            .agg(ratio=('ratio', 'median'), num_peptides=('ratio', len))
            .groupby(level=('uniprot', 'symbol', 'condition'))  
            .agg(ratio=('ratio', 'median'), num_peptides=('num_peptides', 'sum'), ratio_list=('ratio', list))
        )

        # result_df['num_peptides'] = result_df.num_peptides.astype(int)
        result_df = result_df.unstack(level='condition')
        result_df['num_peptides'] = result_df.num_peptides.fillna(0).astype(int)
        # result_df['ratio'] = result_df.ratio.fillna('-')

        def ratios_to_string(ratios, invert=True):
            if not isinstance(ratios, list):
                return
            if invert:
                ratios = [1/r for r in ratios]
            return ', '.join(map(str, ratios))

        result_df['ratio_list'] = result_df.ratio_list.transform(lambda x: x.apply(ratios_to_string))
        result_df['ratio'] = result_df.ratio.rdiv(1)
        
        result_df.columns = result_df.columns.swaplevel()

        sorted_col_multiindex, _ = result_df.columns.sortlevel()
        result_df = result_df[sorted_col_multiindex]

        # result_df.loc[:, pd.IndexSlice[:, 'ratio']] = (result_df
        #     .loc[:, pd.IndexSlice[:, 'ratio']]
        #     .rdiv(1)
        # )

        result_df = (result_df
            .reset_index()
            .sort_values(by=[(self.datasets[0].name, 'ratio'), 'symbol'], ascending=True)
            .fillna('-')
        )

        report_output_name_template = '{}_{}_{}_{{}}.csv'.format(
            report_prefix,
            self.name,
            self._analysis.id
        )

        report_output_path = utils.get_timestamped_report_path(
            report_output_name_template,
            self.output_path
        )

        result_df.to_csv(report_output_path, index=False, encoding='utf-8-sig')
        return result_df

    def filter_report(self, merge_cells=True):
        datasets_to_filter = self.experiment_ids_included

        m = self.data_model

        query = (m
            .select(
                m.id,
                m.experiment,
                m.uniprot,
                m.symbol,
                m.description,
                m.sequence,
                m.clean_sequence,
                m.ratio,
                m.num_ms2,
                m.rsquared,
                m.charge,
                m.meta
            )
            .where(m.experiment_id.in_(datasets_to_filter))
        )

        df = pd.DataFrame.from_records(list(query.dicts()))
        df = self.dataset_class.generate_id(df)

        for dataset in self.datasets:
            df.loc[df.experiment.isin(dataset.experiment_ids_included), 'condition'] = dataset.name

        df = df.set_index('id')

        df = df[[
            '_id',
            'experiment',
            'condition',
            'uniprot',
            'symbol',
            'sequence',
            'meta',
            'num_ms2',
            'rsquared',
            'ratio',
        ]]

        for filtered_out in self._analysis.filters.values():
            for cat, f in filtered_out.items():
                for filter_name, filtered_ids in f.items():
                    df.loc[filtered_ids, f'{cat}.{filter_name}'] = False

        # this should probably be just done in SQL
        # experiment_ids = df.experiment.unique().tolist()
        # q2 = Experiment.select(Experiment.id, Experiment.source_url).where(Experiment.id.in_(experiment_ids))
        # experiments = dict(list(q2.tuples()))
        # df.link = df.apply(lambda x: experiments[x.experiment] + x.link.split('"')[1], axis=1)

        report_output_name_template = 'filter_report_{}_{}_{{}}.xlsx'.format(
            self.name,
            self._analysis.id
        )

        report_output_path = utils.get_timestamped_report_path(
            report_output_name_template,
            self.output_path
        )

        # # df.set_index(['seq_id', 'condition', 'experiment']).sort_index(level=0).to_excel(report_output_path)
        df.set_index(['_id', 'experiment', 'condition']).sort_index(level=0).to_excel(report_output_path, merge_cells=merge_cells)
        return df
    
    def unfiltered_report(self, merge_cells=True):
        m = self.data_model

        query = (m
            .select(
                m.id,
                m.experiment,
                m.uniprot,
                m.symbol,
                m.description,
                m.sequence,
                m.mass,
                m.charge,
                m.rsquared,
                m.ratio,
            )
            .where(
                m.experiment_id.in_(self.experiment_ids_included)
            )
        )

        df = pd.DataFrame.from_records(list(query.dicts()))

        for dataset in self.datasets:
            df.loc[df.experiment.isin(dataset.experiment_ids_included), 'condition'] = dataset.name

        df = df[~df.uniprot.str.startswith('Reverse_')]
        df['description'] = df.description.str.split().str[1:].str.join(' ')

        df = df.set_index(['uniprot', 'symbol', 'description', 'condition', 'experiment']).sort_index(level=0)

        report_output_name_template = 'unfiltered_{}_{}_{{}}.xlsx'.format(
            self.name,
            self._analysis.id
        )

        report_output_path = utils.get_timestamped_report_path(
            report_output_name_template,
            self.output_path
        )

        df.to_excel(report_output_path, encoding='utf-8-sig', merge_cells=merge_cells)
        return df


def analyze_tmt(analysis_params: dict):
    database.init(f'{analysis_params.get("user", config.DEFAULT_DB_NAME)}.db', pragmas=config.pragmas)
    params = AnalysisParams.from_dict(analysis_params)

    analysis = TmtAnalysis(params=params)
    analysis.run()
    return analysis

def analyze_silac(analysis_params: dict):
    database.init(f'{analysis_params.get("user", config.DEFAULT_DB_NAME)}.db', pragmas=config.pragmas)
    params = SilacAnalysisParams.from_dict(analysis_params)

    analysis = SilacAnalysis(params=params)
    analysis.run()
    return analysis



# modules defined in this project
from trifl.utils import uniprot_db_from_fasta, find_all
import trifl.models as m

# installed modules
from bs4 import BeautifulSoup
import requests

# base modules
from collections import defaultdict
from typing import Literal
import functools
import itertools
import pathlib
import re
import io
import csv
import urllib.parse as urlparse


delchars = {ord(c): None for c in map(chr, range(256)) if not c.isalpha()}
delchars_except_mod = {ord(c): None for c in map(chr, range(256)) if not c.isalpha() and not c == '*'}

# fasta_db_path = 'UniProt_Human_Cravattlab_nonredundant2_98id_11-05-2012_reversed.fasta'
# uniprot_db = uniprot_db_from_fasta(fasta_db_path)


def get_dataset_from_url(combined_dta_url: str, dta_select_url: str):
    """Get raw dataset from  given url."""
    res: Response = requests.get(combined_dta_url)
    res.raise_for_status() # raise exception if the url is not found

    dta_res: Response = requests.get(dta_select_url)
    dta_res.raise_for_status()
    ms2_counts, unique_sequences = parse_dtaselect(dta_res.text)

    return parse_peptide(res.text, ms2_counts, unique_sequences)

def get_dta_select_url(combined_dta_url):
    dta_folder_url = urlparse.urljoin(combined_dta_url, 'dta/')
    res = requests.get(dta_folder_url)
    soup = BeautifulSoup(res.text, 'html.parser')
    dta_select_link = soup.find_all('a', string=re.compile('DTASelect'))[0].attrs['href']
    return dta_folder_url + dta_select_link

def parse_dtaselect(dta_file_path: pathlib.Path):
    with dta_file_path.open('r') as f:
        raw_dta = f.read()

    data = raw_dta.splitlines()

    data_start_row = 0
    sequence_column = 0

    # first we find the row where data starts and where the sequence column is
    for index, row in enumerate(data[:100]):
        if row.startswith('Unique'):
            data_start_row = index + 1
            split_row =  row.strip().split('\t')
            sequence_column = split_row.index('Sequence')
            break

    counts = defaultdict(int)
    unique_sequences = set()

    UNIQUE_SYMBOL = '*'

    for row in data[data_start_row:]:
        if row:
            first_character = row[0]
        else:
            continue

        if first_character in ('*', '\t'):
            try:
                sequence = row.split('\t')[sequence_column]

                counts[sequence] = counts[sequence] + 1
                if first_character == UNIQUE_SYMBOL:
                    unique_sequences.add(sequence)
            # reached the end of the file where summary stats lie
            except:
                break

    return (counts, unique_sequences)

class _ParseCombined():
    header_pattern = re.compile('^\d.+')

    def __init__(self):
        pass

    def parse_file(self, filename):
        with open(filename) as combined_dta:
            return self.parse_io(combined_dta)

    def parse_from_string(self, data):
        return self.parse_io(io.StringIO(data))

    def parse_io(self, raw):
        # ditch headings
        headings = raw.readline()
        data = raw.readlines()
        return self.parse_data(data)
        
    def parse_data(self, data):
        groups = []

        for line in data:
            line_info = [ x.strip() for x in re.split('\t+', line.strip()) if len(x.strip()) ]
            # if we've reached a group header
            if self.header_pattern.match(line):
                groups.append(self._extract_header(line_info))
            # if we've reached a subgroup
            else:
                # append peptide to last group
                groups[-1]['peptides'].append(self._extract_line_data(line_info))

        return groups

    def _extract_header(self, line):
        pass

    def _extract_line_data(self, line):
        pass


class ParseProteinCombined(_ParseCombined):
    def _extract_header(self, line):
        return {
            'uniprot': line[1],
            'description': line[2],
            'symbol': line[3],
            'mean_ratio': float(line[4]),
            'peptides': []
        }

    def _extract_line_data(self, line):
        peptide_keys = [
            'sequence', 'mass', 'ratio', 'stats', 'mean', 'noqp', 'run',
            'charge', 'segment', 'link'
        ]
        return dict(zip(peptide_keys, line))

class ParsePeptideCombined(_ParseCombined):
    def _extract_header(self, line):
        return {
            'sequence': line[1],
            'mean_ratio': line[2],
            'peptides': []
        }

    def _extract_line_data(self, line):
        peptide_keys = [
            'uniprot', 'description', 'symbol', 'sequence',
            'mass', 'ratio', 'stats', 'run', 'charge', 'segment', 'link'
        ]

        return dict(zip(peptide_keys, line))


def parse_flatfile(path: pathlib.Path, dta_file_path: pathlib.Path, fasta_db_path: pathlib.Path):
    data = []

    ms2_counts, unique_sequences = parse_dtaselect(dta_file_path)
    fasta_db = m.FastaDatabaseIndex.get_or_create(fasta_db_path)

    delchars_except_mod = {ord(c): None for c in map(chr, range(256)) if not c.isalpha() and not c == '*'}

    with open(path, 'r') as f:
        # skip first line
        f.readline()

        for line in csv.reader(f, delimiter='\t'):
            sequence = line[4]
            clean_sequence = sequence.split('.')[1].translate(delchars)

            data.append({
                'uniprot': line[1],
                'description': line[2],
                'symbol': line[3],
                'sequence': sequence,
                'clean_sequence': clean_sequence,
                'unique_sequence': sequence in unique_sequences,
                # 'residue_number': get_residue_number(line[1], line[4]) if get_residue_number else None,
                'mass': float(line[5]),
                'charge': int(line[6]),
                'segment': int(line[7]),
                'ratio': float(line[8]),
                'num_ms2': ms2_counts.get(sequence, 0),
                'meta': {
                    'intensity': line[9],
                    'peptide_index': line[0],
                    'num_ms2_peaks': line[10].split('/')[0],
                    'num_candidate_peaks': line[10].split('/')[1],
                    'max_light_intensity': line[10].split('/')[2],
                    'light_noise': line[10].split('/')[3],
                    'max_heavy_intensity': line[10].split('/')[4],
                    'heavy_noise': line[10].split('/')[5],
                    'entry': line[12],
                    'link': line[13],
                },
                'rsquared': float(line[11]),
            })

    return data


def parse_protein(raw_dataset: str, ms2_counts: dict, unique_sequences: set):
    parser = ParseProteinCombined()
    data = []

    for group in parser.parse_io(raw_dataset):
        for peptide in group['peptides']:
            full_sequence = peptide['sequence']
            clean_sequence = full_sequence.split('.')[1].translate(delchars)

            data.append({
                'cimage_sequence': '',
                'clean_sequence': clean_sequence,
                'residue': 0,
                'possible_residues': [],
                'num_ms2': ms2_counts.get(full_sequence),
                'unique_sequence': full_sequence in unique_sequences,
                **peptide
            })

    return data

def parse_peptide(raw_dataset: str, ms2_counts: dict, unique_sequences: set):
    parser = ParsePeptideCombined()
    data = []
    
    delchars_except_mod = {ord(c): None for c in map(chr, range(256)) if not c.isalpha() and not c == '*'}

    for group in parser.parse_from_string(raw_dataset):
        cimage_sequence = group['sequence']

        for peptide in group['peptides']:
            full_sequence = peptide['sequence']
            clean_sequence = full_sequence.split('.')[1].translate(delchars)

            data.append({
                'cimage_sequence': cimage_sequence,
                'clean_sequence': clean_sequence,
                'residue': annotate_residue(peptide['uniprot'], full_sequence),
                'possible_residues': get_all_possible_residues(peptide['uniprot'], clean_sequence),
                'num_ms2': ms2_counts.get(full_sequence),
                'unique_sequence': full_sequence in unique_sequences,
                **peptide
            })

    return data


def annotate_residue(uniprot, sequence):
    if uniprot.startswith('Reverse_'):
        return 0

    protein_sequence = uniprot_db[uniprot].seq

    # take sequence between tryptic terminii and remove all modifications
    # except *, which indicates that the residue is labeled
    sequence = sequence.split('.')[1].translate(delchars_except_mod)
    position_in_sequence = sequence.index('*')
    sequence = sequence.replace('*', '')
    sequence_position_in_protein = protein_sequence.find(sequence)
    residue = sequence_position_in_protein + position_in_sequence

    return residue

def get_all_possible_residues(uniprot, sequence, residue='K'):
    if uniprot.startswith('Reverse_'):
        return [0]

    protein_sequence = uniprot_db[uniprot].seq
    sequence_position_in_protein = protein_sequence.find(sequence)

    positions_in_sequence = find_all(sequence, residue)
    positions_in_protein = [sequence_position_in_protein + x + 1 for x in positions_in_sequence]

    return positions_in_protein



from trifl.utils import (
    ResidueLocator,
    get_first_float_in_string as find_float
)

import numpy as np

import csv
import io
import re

delchars = {ord(c): None for c in map(chr, range(256)) if not c.isalpha()}
delchars_except_mod = {ord(c): None for c in map(chr, range(256)) if not c.isalpha() and not c == '*'}


class ParseCensus():
    def __init__(self):
        self.protein_headers = []
        self.peptide_headers = []

    def parse_file(self, filename):
        headers = []

        with open(filename) as f:
            where_was_i = 0
            line = f.readline()
            # collect the headers first
            while line:
                if line.startswith('H'):
                    headers.append(line.strip().split('\t'))
                    where_was_i = f.tell()
                    line = f.readline()
                else:
                    # roll it back so we can pick up where we left
                    # off and parse the data
                    f.seek(where_was_i)
                    break

            self.protein_headers = headers[-2][2:]
            self.peptide_headers = headers[-1][2:]

            # and now parse the data with a real parser
            reader = csv.reader(f, delimiter='\t')
            return self.parse_io(reader)

    def parse_io(self, reader):
        data = []

        for row in reader:
            if row[0] == 'P':
                protein = dict(zip(self.protein_headers, row[1:]))
                protein['peptides'] = []
                data.append(protein)
            elif row[0] == 'S':
                peptide = dict(zip(self.peptide_headers, row[1:]))
                protein['peptides'].append(peptide)

        return data

def parse_census(filename, fasta_db_path):
    parser = ParseCensus()
    raw = parser.parse_file(filename)

    data = []

    oxidized_methionine_regex = re.compile(f'(M)\([\d\.]+\)')
    symbol_regex = re.compile(r'GN=(\w+)')

    for item in raw:
        symbol = symbol_regex.search(item['DESCRIPTION'])
        protein = {
            'uniprot': item['LOCUS'],
            'description': ' '.join(item['DESCRIPTION'].split('=')[0].split()[:-1]),
            'symbol': symbol[1] if symbol else ''
        }

        for peptide in item['peptides']:
            sequence = peptide['SEQUENCE']
            sequence = oxidized_methionine_regex.sub(r'\1#', sequence)
            # sequence = re.sub(r'\([\d\.]+\)', '*', peptide['SEQUENCE'])
            intensities = {find_float(k): int(v) for k, v in peptide.items() if k.startswith('m/z')}
            
            data.append({
                'sequence': sequence,
                'clean_sequence': sequence.split('.')[1].translate(delchars),
                'unique_sequence': peptide['UNIQUE'] == 'U',
                'channel_intensities': list(intensities.values()),
                'channel_masses': list(intensities.keys()),
                'scan_num': int(peptide['ScanNum']),
                'tmt_purity': np.nan_to_num(float(peptide['TMT_purity'])),
                'filename': peptide['Filename'],
                # 'meta': {
                #     'raw': peptide
                # },
                **protein
            })

    return data

def parse_census_residue(filename, fasta_db_path, residue):
    parser = ParseCensus()
    raw = parser.parse_file(filename)

    data = []

    locator = ResidueLocator(fasta_db_path, residue=residue)

    reactive_residue_regex = re.compile(f'({residue})\([\d\.]+\)')
    oxidized_methionine_regex = re.compile(f'(M)\([\d\.]+\)')

    for item in raw:
        protein = {
            'uniprot': item['LOCUS'],
            'description': '',
            'symbol': ''
        }

        for peptide in item['peptides']:
            sequence = reactive_residue_regex.sub(r'\1*', peptide['SEQUENCE'])
            sequence = oxidized_methionine_regex.sub(r'\1#', sequence)
            # sequence = re.sub(r'\([\d\.]+\)', '*', peptide['SEQUENCE'])
            intensities = {find_float(k): int(v) for k, v in peptide.items() if k.startswith('m/z')}

            data.append({
                'sequence': sequence,
                'clean_sequence': sequence.split('.')[1].translate(delchars),
                'unique_sequence': peptide['UNIQUE'] == 'U',
                'residue': locator.locate_residue(protein['uniprot'], sequence),
                'possible_residues': locator.get_all_possible_residues(protein['uniprot'], sequence),
                'channel_intensities': list(intensities.values()),
                'channel_masses': list(intensities.keys()),
                # 'meta': {
                #     'raw': peptide
                # },
                **protein
            })

    return data



from .census import parse_census
from .cimage import (
    parse_dtaselect,
    parse_flatfile as parse_cimage_flatfile,
    parse_peptide as parse_cimage_peptide,
    parse_protein as parse_cimage_protein,
)


tmt_parsers = {
    'census': parse_census
}

cimage_parsers = {
    'flatfile': parse_cimage_flatfile,
    'peptide': parse_cimage_peptide,
    'protein': parse_cimage_protein
}



from Bio import SeqIO
from bs4 import BeautifulSoup
import pandas as pd
import numpy as np
import requests

from typing import List, Iterable
import urllib.parse as urlparse
import datetime
import dataclasses
import itertools
import json
import hashlib
import inspect
import subprocess
import pathlib
import base64
import difflib
import re

_words_only = re.compile('[\W_]+', re.UNICODE)

class ResidueLocator:
    delchars = {ord(c): None for c in map(chr, range(256)) if not c.isalpha()}

    def __init__(self, fasta_db_path, residue, mod_symbol='*'):
        self.uniprot_db = uniprot_db_from_fasta(fasta_db_path)
        self.symbol = mod_symbol
        self.delchars_except_mod = {ord(c): None for c in map(chr, range(256)) if not c.isalpha() and not c == self.symbol}
        self.residue = residue
    
    def locate_residue(self, uniprot: str, sequence: str) -> int:
        if uniprot.startswith('Reverse_'):
            return 0

        protein_sequence = self.uniprot_db[uniprot].seq

        # take sequence between tryptic terminii and remove all modifications
        # except *, which indicates that the residue is labeled
        try:
            sequence = sequence.split('.')[1].translate(self.delchars_except_mod)
            position_in_sequence = sequence.index(self.symbol)
            sequence = sequence.replace(self.symbol, '')
            sequence_position_in_protein = protein_sequence.find(sequence)
            residue = sequence_position_in_protein + position_in_sequence
        except ValueError:
            # handle unmodified peptides
            residue = 0

        return residue

    def get_all_possible_residues(self, uniprot: str, sequence: str) -> int:
        if uniprot.startswith('Reverse_'):
            return [0]

        sequence = sequence.translate(self.delchars)
        protein_sequence = self.uniprot_db[uniprot].seq
        sequence_position_in_protein = protein_sequence.find(sequence)

        positions_in_sequence = find_all(sequence, self.residue)
        positions_in_protein = [sequence_position_in_protein + x + 1 for x in positions_in_sequence]

        return positions_in_protein


def uniprot_db_from_fasta(path):
    db = list(SeqIO.parse(str(path), 'fasta'))
    uniprot_db = {}

    for item in db:
        try:
            if not 'Reverse_' in item.id:
                # Reversed entries can have the same id as the parent entry, and thus would
                # lead to overwriting the correct sequence for a given uniprot
                # with the reverse sequence
                _id = item.id.split('|')[1]
                uniprot_db[_id] = item
        except IndexError:
            pass

    return uniprot_db

def get_git_revision_hash():
    # credit to Yuji Tomita: https://stackoverflow.com/a/21901260/383744
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip()

def hash_fn_source(fn):
    return hashlib.md5(inspect.getsource(fn).encode()).hexdigest()

def get_short_dataframe_hash(df, hash_length=4):
    h = hashlib.sha1(pd.util.hash_pandas_object(df, index=True).values)
    le_hash = _words_only.sub('', base64.b64encode(h.digest()).decode())
    return le_hash[:hash_length]

class JSONSetEncoder(json.JSONEncoder):
    def default(self, obj):
       if isinstance(obj, set):
          return list(obj)
       return json.JSONEncoder.default(self, obj)

def get_datestamp():
    return datetime.datetime.now().strftime('%Y%m%d%H%M')

def get_timestamped_report_path(template, data_output_path, datestamp=None):
    if not datestamp:
        datestamp = get_datestamp()
    return data_output_path.joinpath(template.format(datestamp))

def get_newest_file(base_path: pathlib.Path, glob: str ='*') -> pathlib.Path:
    newest_file_path = next(iter(sorted(
        (f for f in base_path.glob(glob)),
        key=lambda x: x.stat().st_ctime, reverse=True
    )), None)

    return newest_file_path

def pretty_residues(residue_list):
    return ', '.join(map(
        'C{}'.format,
        sorted(set(residue_list))
    ))

def get_longest_overlapping_sequence(sequences: List[str]) -> str:
    sequences = list(set(sequences))

    if len(sequences) == 1:
        return sequences[0]

    d = difflib.SequenceMatcher(a=sequences[0], b=sequences[1])
    match = d.find_longest_match(0, len(sequences[0]), 0, len(sequences[1]))

    overlap_seq = sequences[0]
    d = difflib.SequenceMatcher(a=overlap_seq)

    for seq in sequences[1:]:
        d.set_seq2(seq)
        match = d.find_longest_match(0, len(overlap_seq), 0, len(seq))
        overlap_seq = seq[match[1]:match[1] + match[2]]
        d.set_seq1(overlap_seq)

    return overlap_seq

# https://stackoverflow.com/a/4665027/383744
def find_all(a_str, sub):
    start = 0
    while True:
        start = a_str.find(sub, start)
        if start == -1: return
        yield start
        start += len(sub) # use start += 1 to find overlapping matches

def partition(pred, iterable):
    'Use a predicate to partition entries into true entries and false entries'
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    return list(filter(pred, t2)), list(itertools.filterfalse(pred, t1))

def compare_dfs(before, after):
    combined  = pd.concat([before, after])
    diffs = combined.drop_duplicates(keep=False)
    return diffs

def compare_output_excels(before, after):
    return(compare_dfs(
        pd.read_excel(before, index_col=0),
        pd.read_excel(after, index_col=0),
    ))


find_float_regex = re.compile('\d+(?:\.\d+)?')

def get_first_float_in_string(string: str) -> float:
    # https://stackoverflow.com/a/26192945/383744
    return float(find_float_regex.findall(string)[0])

def get_file_hash(path: str, chunk_size: int = 8192) -> str:
    with open(path, 'rb') as f:
        file_hash = hashlib.sha256()
        while chunk := f.read(8192):
            file_hash.update(chunk)

    return file_hash.hexdigest()

def combine_hashes(hashes: Iterable[str]) -> str:
    return hashlib.sha1('|'.join(hashes))

def slash_join(*args):
    '''
    Joins strings to form an URL

    From: https://codereview.stackexchange.com/a/175423/2397
    '''
    return '/'.join(arg.strip('/') for arg in args)


def flatten(list_of_lists: Iterable[Iterable]) -> list:
    '''Yield items from any nested iterable; see Reference.
    
    https://stackoverflow.com/a/40857703/383744
    '''
    for x in list_of_lists:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x

def get_dta_select_url(dta_folder_url: str):
    res = requests.get(dta_folder_url)
    soup = BeautifulSoup(res.text, 'html.parser')
    dta_select_regex = 'DTASelect-filter_.+\.txt'
    dta_select_link = soup.find_all('a', string=re.compile(dta_select_regex))[0].attrs['href']
    return f'{dta_folder_url}/{dta_select_link}'

def download_cimage(
    base_url: str,
    download_path: pathlib.Path,
    cimage_folder_name: str,
    dta_folder_name: str
) -> bool:
    flatfile_name = 'output_rt_10_sn_2.5.to_excel.txt'
    combined_dta_name =  f'combined_{dta_folder_name}.txt'

    base = slash_join(base_url, cimage_folder_name)
    dta_folder_url = slash_join(base, dta_folder_name)
    dta_select_url = get_dta_select_url(dta_folder_url)
    combined_dta_url = slash_join(base, combined_dta_name)
    flatfile_url = slash_join(dta_folder_url, 'output', flatfile_name)

    download_path.mkdir(exist_ok=True, parents=True)

    with download_path.joinpath(combined_dta_name).open('w') as f:
        res: requests.Response = requests.get(combined_dta_url)
        res.raise_for_status() # raise exception if the url is not found
        f.write(res.text)

    with download_path.joinpath(dta_select_url.split('/')[-1]).open('w') as f:
        res: requests.Response = requests.get(dta_select_url)
        res.raise_for_status() # raise exception if the url is not found
        f.write(res.text)

    with download_path.joinpath(flatfile_name).open('w') as f:
        res: requests.Response = requests.get(flatfile_url)
        res.raise_for_status() # raise exception if the url is not found
        f.write(res.text)

    return True



from trifl.filters.shared import Filter, FilterLevel
import trifl.utils as utils
import trifl.models as m
import trifl.config as config

from pandas.core.common import flatten
import pandas as pd
import numpy as np

from typing import List, Dict
from abc import ABC, abstractmethod
from collections import defaultdict
import pathlib


class _Dataset(ABC):
    
    @classmethod
    @abstractmethod
    def generate_id(cls, df):
        '''Generate a unique identifier per item.'''

    @abstractmethod
    def as_df(self, required_columns):
        '''Method for creating a pandas dataframe from the dataset.'''

    @abstractmethod
    def apply_filters(self, filters):
        '''Apply all filters'''

    @property
    def experiment_ids_included(self):
        return [e.id for e in self.experiments]

    def get_required_columns(self):
        filter_requirements = set()

        for level, fs in self.filters.items():
            filter_requirements.update(flatten(f.requires.database_fields for f in fs))

        requirements = self.base_requirements.union(filter_requirements)

        return [getattr(self.data_model, r) for r in requirements]

    def as_dict(self):
        return {
            'name': self.name,
            'experiments': self.experiment_ids_included
        }

class SilacDataset(_Dataset):
    base_requirements = {'id', 'experiment', 'uniprot', 'ratio', 'clean_sequence'}
    data_model = m.SilacData
    experiment_model = m.SilacExperiment

    def __init__(self,
        name: str,
        filters: Dict[FilterLevel, List[Filter]],
        replicate_paths: List[pathlib.Path],
        fasta_db_path: str,
        parser: str,
        base_url: str,
        combine_level: str = config.DEFAULT_SILAC_COMBINE_LEVEL,
        dta_folder_name: str = config.DEFAULT_DTA_FOLDER_NAME,
        meta = None
    ) -> None:
        self.name = name
        self.filters = filters
        self.replicate_paths = replicate_paths
        self.fasta_db_path = fasta_db_path
        self.meta = meta
        self.parser = parser
        self.base_url = base_url
        self.combine_level = combine_level
        self.dta_folder_name = dta_folder_name
        self.experiments = self._get_or_create_experiments()

    @classmethod
    def generate_id(cls, df):
        df['_id'] = df['uniprot']
        return df

    def as_df(self, required_columns):
        Data = self.data_model

        query = (Data
            .select(*required_columns)
            .where(Data.experiment_id.in_(self.as_dict()['experiments']))
        )

        df = pd.DataFrame.from_records(list(query.dicts()))
        df = df.set_index('id')

        return df

    def apply_filters(self):
        df = self.as_df(self.get_required_columns())
        df = self.generate_id(df)

        # replace zero with NaN to avoid counting in mean and stats calc
        df = df.replace(0, np.NaN)

        filtered_out = {}

        if 'individual' in self.filters:
            df, filtered_out['individual'] = self.apply_individual_filters(df)

        if 'experiment' in self.filters:
            df, filtered_out['experiment'] = self.apply_experiment_filters(df)

        if 'aggregate' in self.filters:
            df, filtered_out['aggregate'] = self.apply_aggregate_filters(df)

        df = df[~df.exclude]

        return df, filtered_out

    def apply_individual_filters(self, df):
        filter_category = 'individual'
        filtered_out = {}

        for f in self.filters.get('individual'):
            reqs = f.requires.dataframe_columns

            if isinstance(reqs, str):
                reqs = [reqs]

            f_colname = f'{filter_category}.{f.name}'

            df[f_colname] = df[reqs].apply(
                lambda x: f.fn(*(x[r] for r in reqs)),
                axis=1,
            )

            filtered_ids = set(df[~df[f_colname]].index.values.tolist())
            filtered_out[f.name] = filtered_ids

        df['exclude'] = ~df.filter(regex=f'^{filter_category}\.').all(axis=1)

        return df, filtered_out

    def apply_experiment_filters(self, df):
        filter_category = 'experiment'
        filtered_out = {}

        df = df[~df.exclude].copy()

        for f in self.filters.get('experiment'):
            f_colname = f'{filter_category}.{f.name}'

            grouped = df[~df.exclude].groupby(['_id', 'experiment'])
        
            df[f_colname] = grouped[f.requires.dataframe_columns].apply(
                lambda x: pd.Series(f.fn(x.tolist()), index=x.index)
            )

            filtered_out[f.name] = set(df[df[f_colname] == False].index.values.tolist())

            # make sure to exclude items that were filtered out for next filter
            df.loc[df[f_colname] == False, 'exclude'] = True

        return df, filtered_out


    def apply_aggregate_filters(self, df):
        filter_category = 'aggregate'
        filtered_out = defaultdict(set)

        df = df[~df.exclude].copy()

        for f in self.filters.get('aggregate'):
            f_colname = f'{filter_category}.{f.name}'

            # note that this implies our filters can only access the ratio column
            # instead could apply here and avoid two groupbys
            grouped = df[~df.exclude].reset_index().groupby(['_id', 'experiment']).agg({
                'ratio': 'median',
                'clean_sequence': list,
                'id': list
            })

            for _, g in grouped.groupby(level='_id'):
                data = g[f.requires.dataframe_columns].values.tolist()
                ids = g.id.values
                ids_to_filter_out = [x[1] for x in zip(f.fn(data), ids) if not x[0]]

                for y in ids_to_filter_out:
                    filtered_out[f.name].update(y)

            df.loc[filtered_out[f.name], f_colname] = False
            df.loc[filtered_out[f.name], 'exclude'] = True

        return df, filtered_out

    def _get_or_create_experiments(self):
        return [self.experiment_model.get_or_create(
            name=self.name,
            cimage_folder_path=path,
            fasta_db_path=self.fasta_db_path,
            base_url=self.base_url,
            parser=self.parser,
            dta_folder_name=self.dta_folder_name,
        ) for path in self.replicate_paths]

class _TmtDataset(_Dataset):
    base_requirements = {'id', 'experiment', 'uniprot', 'channel_intensities'}

    def as_df(self, required_columns):
        Data = self.data_model

        query = (Data
            .select(*required_columns)
            .where(Data.experiment_id.in_(self.as_dict()['experiments']))
        )

        df = pd.DataFrame.from_records(list(query.dicts()))
        df = df.set_index('id')

        for channel in self.channel_layout:
            channel_num, channel_name = next(iter(channel.items()))
            new_col_name = f'{channel_name}.intensity'
            # append number to column to distinguish duplicates
            new_col_name = f'{new_col_name}_{len(df.filter(regex=new_col_name).columns)}'
            df[new_col_name] = df.channel_intensities.str.get(channel_num - 1)
        df = df.drop(columns=['channel_intensities'])

        return df


class TmtDataset(_TmtDataset):
    data_model = m.TmtData
    experiment_model = m.TmtExperiment

    def __init__(self,
        name: str,
        filters: Dict[FilterLevel, List[Filter]],
        replicate_paths: List[str],
        channel_layout: dict,
        control_channels: str,
        fasta_db_path: str,
        meta = None
    ) -> None:
        self.name = name
        self.filters = filters
        self.replicate_paths = replicate_paths
        self.channel_layout = channel_layout
        self.control_channels = control_channels
        self.fasta_db_path = fasta_db_path
        self.meta = meta
        self.experiments = self._get_or_create_experiments()

    @classmethod
    def generate_id(cls, df):
        df['_id'] = df['uniprot']
        return df

    def _get_or_create_experiments(self):
        return [self.experiment_model.get_by_file_hash(path) or self.experiment_model.create_from_census(
            name=self.name,
            census_file_path=path,
            fasta_db_path=self.fasta_db_path
        ) for path in self.replicate_paths]

    def apply_filters(self):
        df = self.as_df(self.get_required_columns())
        df = self.generate_id(df)

        control_cols = df.filter(regex=f'{self.control_channels}.intensity')
        data_cols = df.filter(regex=f'.+\.intensity')

        percentages = data_cols.div(control_cols.mean(axis=1), axis=0).mul(100, axis=0)
        percentages = percentages.rename(columns={
            c: c.replace('intensity', 'percent_of_control')
            for c in df.filter(regex='\.intensity').columns
        })

        df = df.join(percentages)

        # note that this is done here since attrs get overwritten when two
        # dataframes (df + percentages) are joined
        df.attrs['control_intensities'] = list(control_cols)
        df.attrs['channel_intensities'] = list(data_cols)

        # replace zero with NaN to avoid counting in mean and stats calc
        # df = df.replace(0, np.NaN)

        filtered_out = {}

        if 'individual' in self.filters:
            df, filtered_out['individual'] = self.apply_individual_filters(df)

        if 'experiment' in self.filters:
            df, filtered_out['experiment'] = self.apply_experiment_filters(df)

        if 'aggregate' in self.filters:
            df, filtered_out['aggregate'] = self.apply_aggregate_filters(df)

        df = df[~df.exclude]

        return df, filtered_out

    def apply_individual_filters(self, df):
        filtered_out = {}

        for f in self.filters.get('individual'):
            reqs = f.requires.dataframe_columns
            filter_colname = f'individual.{f.name}'

            try:
                df[filter_colname] = f.fn(df[df.attrs.get(reqs, reqs)].values.T)
            except:
                df[filter_colname] = df[df.attrs.get(reqs, reqs)].apply(f.fn)

            filtered_out[f.name] = set(df[~df[filter_colname]].index.values.tolist())

        df['exclude'] = ~df.filter(regex=f'^individual\.').all(axis=1)

        return df, filtered_out

    def apply_experiment_filters(self, df):
        filtered_out = {}

        df = df[~df.exclude].copy()

        grouped = df.groupby(['_id', 'experiment']).agg({
            'uniprot': 'first',
            'experiment': list,
            **{c: 'mean' for c in df.filter(regex='percent_of_').columns}
        })

        for f in self.filters.get('experiment'):
            reqs = f.requires.dataframe_columns
            filter_colname = f'experiment.{f.name}'

            try:
                df[filter_colname] = f.fn(df[df.attrs.get(reqs, reqs)].values.T)
            except:
                df[filter_colname] = df[df.attrs.get(reqs, reqs)].apply(f.fn)

            filtered_out[f.name] = set(df[~df[filter_colname]].index.values.tolist())

        return df, filtered_out

    def apply_aggregate_filters(self, df):
        filtered_out = {}

        df  = df[~df.exclude].copy()

        for f in self.filters.get('aggregate'):
            reqs = f.requires.dataframe_columns
            filter_colname = f'aggregate.{f.name}'

            grouped = df[~df.exclude].reset_index()

        return df, filtered_out


# class TmtIsotopDataset(_Dataset):
#     def __init__(self,
#         name: str,
#         replicate_paths: List[str],
#         channel_layout: dict,
#         control_channels: str,
#         fasta_db_path: str,
#         reactive_residue: str,
#         meta = None
#     ) -> None:
#         self.name = name
#         self.replicate_paths = replicate_paths
#         self.channel_layout = channel_layout
#         self.control_channels = control_channels
#         self.fasta_db_path = fasta_db_path
#         self.reactive_residue = reactive_residue
#         self.meta = meta
#         self.experiments = self._get_or_create_experiments()

#     def _get_or_create_experiments(self):
#         return [m.TmtExperiment(
#             self.name,
#             path,
#             self.channel_layout,
#             reactive_residue=self.reactive_residue,
#             fasta_db_path=self.fasta_db_path
#         ) for path in self.replicate_paths]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trifl', 'trifl.filters', 'trifl.parse']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'biopython>=1.78,<2.0',
 'click>=8.0.1,<9.0.0',
 'numpy>=1.20.3,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'peewee>=3.14.4,<4.0.0',
 'requests>=2.25.1,<3.0.0',
 'scipy>=1.6.3,<2.0.0']

setup_kwargs = {
    'name': 'trifl',
    'version': '0.1.4',
    'description': 'TRIFL is a data filtration library for MS-proteomics experiments.',
    'long_description': None,
    'author': 'Radu Suciu',
    'author_email': 'radusuciu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .analysis import (
    analyze_tmt,
    analyze_silac
)


import yaml

def analyze(analysis_params_filepath: str, user: str):
    with open(analysis_params_filepath, 'r') as f:
        analysis_params = yaml.safe_load(f)

    analysis_params['user'] = user

    analysis_fn_map = {
        # 'tmt_isotop': analyze_tmt_isotop,
        'tmt': analyze_tmt,
        'silac': analyze_silac
    }

    analysis_fn = analysis_fn_map[analysis_params['type']]

    return analysis_fn(analysis_params), analysis_params



DEFAULT_DB_NAME = 'trifl'
DEFAULT_REACTIVE_RESIDUE = 'C'
DEFAULT_CIMAGE_PARSER = 'flatfile'
DEFAULT_SILAC_COMBINE_LEVEL = 'protein'
DEFAULT_DTA_FOLDER_NAME = 'dta'
DEFAULT_REPORT_PREFIX = 'report'


pragmas = [
    ('journal_mode', 'wal'),
    ('cache_size', -1000 * 32)
]



from trifl.exceptions import FilterNotFoundException
import trifl.utils as utils

from collections import namedtuple
from typing import List, Literal, Callable, Union, Optional
from dataclasses import dataclass
import functools

FilterFunction = Callable[..., Union[bool, List[bool]]]
FilterLevel = Literal['individual', 'experiment', 'aggregate']


@dataclass
class Requirement:
    database_fields: Union[str, List[str]]
    dataframe_columns: Union[str, List[str]]

@dataclass
class Filter:
    fn: FilterFunction
    name: str
    requires: Requirement
    level: FilterLevel

    @classmethod
    def from_dict(cls, user_defined_filter: Union[str, dict], filters_module):
        f_kwargs = {}
        f_args = []

        if isinstance(user_defined_filter, dict):
            f, f_kwargs = next(iter(user_defined_filter.items()))

            if not isinstance(f_kwargs, dict):
                f_kwargs = {}
                f_args = f_kwargs
        else:
            f = user_defined_filter

        filter_obj = getattr(filters_module, f)

        if filter_obj is None:
            raise FilterNotFoundException
        
        filter_obj.fn = functools.partial(filter_obj.fn, *f_args, **f_kwargs)
        return filter_obj

    @property
    def dataframe_reqs(self):
        reqs = self.requires.dataframe_columns
        return reqs if isinstance(reqs, list) else [reqs]

def filter_fun(requires: Union[Requirement, str], level: Union[FilterLevel, List[FilterLevel]]):
    if isinstance(requires, str):
        requires=Requirement(
            database_fields=requires,
            dataframe_columns=requires
        )
    def inner_decorator(fn: FilterFunction) -> Filter:
        return Filter(fn=fn, name=fn.__name__, requires=requires, level=level)
    return inner_decorator

f = filter_fun
r = Requirement

@f(requires='sequence', level='individual')
def is_not_half_tryptic(full_sequence: str) -> bool:
    return full_sequence[0] in ['K','R','-'] and (full_sequence[-3] in ['K','R'] or full_sequence[-1] == '-')

@f(requires='sequence', level='individual')
def is_not_terminal(full_sequence: str) -> bool:
    return '-' not in full_sequence

@f(requires='uniprot', level='individual')
def is_not_reverse(uniprot: str) -> bool:
    return not uniprot.startswith('Reverse_')

@f(requires='clean_sequence', level='individual')
def is_not_miscleaved(clean_sequence: str) -> bool:
    seq = clean_sequence[:-1]
    return not('K' in seq or 'R' in seq)

@f(requires='clean_sequence', level='individual')
def is_not_miscleaved_more_than_once(clean_sequence: str) -> bool:
    seq = clean_sequence[:-1]
    return (seq.count('K') + seq.count('R')) < 2

@f(requires='unique_sequence', level='individual')
def is_unique_sequence(unique_sequence: bool) -> bool:
    return unique_sequence

@f(requires='unique_sequence', level='individual')
def is_non_unique_sequence(unique_sequence: bool) -> bool:
    return not unique_sequence

@f(requires='clean_sequence', level=['experiment', 'aggregate'])
def min_num_unique_sequences(clean_sequences: List[str], min_unique_sequences: int=2) -> bool:
    no_filter = [True] * len(clean_sequences)
    filter_all = [False] * len(clean_sequences)
    return no_filter if len(set(utils.flatten(clean_sequences))) >= min_unique_sequences else filter_all
